/**
 * $Id: Actor.java,v 1.61 2003/09/09 22:25:13 deselt Exp $
 *
 * This is the main Actor class.
 */

package salsa.language;

import java.io.*;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.util.Vector;
import java.util.Hashtable;

import salsa.language.exceptions.*;
import salsa.messaging.*;
import salsa.messaging.exceptions.*;
import salsa.naming.*;
import salsa.naming.exceptions.*;

import salsa.resources.StandardOutput;
import salsa.resources.StandardError;

public class Actor extends Thread implements java.io.Serializable {
	public UAN __uan;
	public UAL __ual;

	public Vector mailbox = new Vector();

	protected boolean migrated = false;
	protected boolean stayAlive = true;
	protected TokenInput currentContinuation = null;

	public transient TransportService transportService;
	public transient NamingService namingService;
	public transient TheaterService theaterService;

	public static StandardOutput standardOutput = StandardOutput.getInstance();
	public static StandardError standardError = StandardError.getInstance();

	//this is a vector of all the messages generated by a message
	//processed by this actor.  These are all sent when the message
	//has completed being processed
	public Vector messages = new Vector();

	public Actor() {
		//get the implementations for the required services for an Actor
		transportService = ServiceFactory.getTransport();
		namingService = ServiceFactory.getNaming();
		theaterService = ServiceFactory.getTheater();

		__uan = null;
		__ual = theaterService.generateUAL();

		setName(getClass().getName());
		start();
	}

	/** Die kills the actor when its mailbox has been
	  * emptied
	  */
	public void die()	{ stayAlive = false; }
	/** Destroy immediately kills the actor without processing any
	  * more messages
	  */
	public void destroy()	{ migrated = true; }

	public UAN getUAN()	{ return __uan; }
	public UAL getUAL()	{ return __ual; }


	private Hashtable unresolvedTokens = new Hashtable();
	private Vector pendingTokens = new Vector();

	public synchronized void resolveToken(String id, Object value) {
		TokenOutput target = (TokenOutput)unresolvedTokens.get(id);

		if (target != null) {
			target.replace(value);

			if (target.requiredValues < 1) unresolvedTokens.remove(id);
		} else {
			//The token has reached the actor before the message it was intended for,
			//add this to a vector of tokens waiting for messages.

			pendingTokens.add( new PendingToken(id, value) );
//			System.err.println("target token did not exist for:" + id + " with value: " + value);
		}
	}

	public void addUnresolvedToken(TokenOutput unresolvedToken) {
		unresolvedTokens.put( unresolvedToken.getId(), unresolvedToken );
	}

	public synchronized void send(Message message) {
		RunTime.receivedMessage();

		if (((__uan != null) && (theaterService.getEntry(__uan) == this)) ||
		    ((__ual != null) && (theaterService.getEntry(__ual) == this))) {
			if (message.getRequiredTokens() > 0) {
				Object[] arguments = message.getArguments();
				for (int i = 0; i < arguments.length; i++) {
					if (arguments[i] instanceof TokenOutput) {
						addUnresolvedToken( (TokenOutput)arguments[i] );
					}
				}
				if (message.getInput() instanceof TokenOutput) {
					addUnresolvedToken( (TokenOutput)message.getInput() );
				}

				//try and resolve the pending tokens
				for (int i = 0; i < pendingTokens.size(); i++) {
					PendingToken pendingToken = (PendingToken)pendingTokens.get(i);
					TokenOutput target = (TokenOutput)unresolvedTokens.get( pendingToken.id );

					if (target != null) {
						target.replace( pendingToken.value );
						if (target.requiredValues < 1) unresolvedTokens.remove( pendingToken.id);
						pendingTokens.remove(i);
						i--;
					}
				}

			} else {
				if (message.getProperty() != null && message.getProperty().equals("priority")) {
					mailbox.insertElementAt(message, 0);
				} else {
					mailbox.addElement(message);
				}
				notify();
			}
		} else {
			message.getTarget().send(message);
			notify();
		}
	}

	public synchronized Message getMessage() {
		if (mailbox.isEmpty()) {
			try {
				wait();      // The lock of Mailbox means it is empty.
			} catch (InterruptedException ie){
				System.err.println("Getting a new message:"+ie);
			}
		}
		Message nextMessage = (Message)mailbox.firstElement();
		mailbox.removeElementAt(0);

		return nextMessage;
	}

	public void run() {
		while ((!migrated) && (stayAlive || !mailbox.isEmpty()) ) {
			Message nextMessage = getMessage();

			currentContinuation = nextMessage.getOutput();
			process(nextMessage);
			currentContinuation = null;
		}
	}

	public String toString() {
		return getClass().getName()+":"+super.toString();
	}


	/**
	 * Binds this object to a UAN and UAL so it can be located later.
	 *
	 * @param uan The universal actor name to bind this actor to.
	 * @param ual The universal actor locator to bind this actor to.

	 */
	public void bind(String uan, String ual) throws MalformedUANException, MalformedUALException {
		bind(new UAN(uan), new UAL(ual));
	}

	public synchronized void bind(UAN uan, UAL ual) {
		//bind this actor, then migrate it to the appropriate target.
		bind(uan);
		migrate(ual);
	}

	/**
	 * This method binds the actor to the specified name and adds the actor to
	 * an unspecified location at the current theater. An on-the-fly theater is created
	 * if one does not already exist.
	 * 
	 * @author stepha
	 * @param uan or ual
	 */
	public void bind( String name ) throws MalformedUANException {
		bind( new UAN( name ) );
	}
	
	public synchronized void bind( UAN uan ) {
		this.__uan = uan;

		theaterService.setEntry(__ual, __uan);
		theaterService.setEntry(__uan, this);

		// Update the naming service.
		namingService.add( __uan, __ual );
		RunTime.receivedUniversalActor();
	}

	public synchronized void reName ( String identifier ) {
		UAL ual = new UAL( "rmsp://" + __ual.getHost() + ":" + __ual.getPort() + identifier );

		if (__uan != null) {
			theaterService.setEntry(ual, __uan);
		} else {
			theaterService.setEntry(__ual, ual);
			theaterService.setEntry(ual, this);
		}

		this.__ual = ual;

		// Update the naming service.
		if (__uan != null) namingService.add( __uan, __ual );

		RunTime.receivedUniversalActor();
	}


        public boolean equals(Object o) {
                if (o instanceof ActorReference) {
                        ActorReference ref = (ActorReference)o;
                        //make sure all the references on these actors
                        //are correct
			ref.getActor();
                        
                        if (__uan != null && ref.uan != null) return __uan.equals(ref.uan);
                        else if (__ual != null && ref.ual != null) return __ual.equals(ref.ual);
                } if (o instanceof Actor) {
                        Actor ref = (Actor)o;

                        if (__uan != null && ref.getUAN() != null) return __uan.equals(ref.getUAN());
                        else if (__ual != null && ref.getUAL() != null) return __ual.equals(ref.getUAN());
                }
                return false;
        }

	/**
	 * This method moves the actor from one theater to another.
	 */
	public void migrate(String newUAL) throws MalformedUALException, MigrationException {
		migrate(new UAL(newUAL));
	}

	public synchronized void migrate(UAL ual) throws MigrationException {
		// Update the name server.
		if (__uan != null) {
			namingService.update(__uan, ual);
		}

		this.__ual = ual;
		try {
			// Perform the migration.

			// The order of upsate and migrate has been switched. This might create a bug
			transportService.migrate(this, __ual);

			if (theaterService.getEntry( __uan ) == this) {
				//If this actor was migrated to the same theater,
				//the theater's entry for this __uan will return
				//that serialized actor, which is not == to this one.
				//this allows actors to migrate/bind to the same theater
				//and not have the system break.
				theaterService.setEntry( __uan, __ual );
			}

		} catch (SalsaException e) {
			if (__uan != null) namingService.update( __uan, __ual );
		}
			
		migrated = true;
	}



	/** The following are used to serialize and deserialize an Actor
	 */
	private void writeObject(java.io.ObjectOutputStream out) throws IOException {
		out.defaultWriteObject();
		out.flush();
	}

	private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {
		in.defaultReadObject();
		
		// Get service references from the ServiceFactory.
		transportService = ServiceFactory.getTransport();
		namingService = ServiceFactory.getNaming();
		theaterService = ServiceFactory.getTheater();

		theaterService.setEntry(__uan, this);
		theaterService.setEntry(__ual, __uan);

		standardOutput = StandardOutput.getInstance();
		standardError = StandardError.getInstance();

		setName(getClass().getName());
	}



	/**
		The process method is responsible for determining which
		child class method should be invoked for a specific message.
		The process method first invokes the chosen method, updates any
		tokens and then sends any continuation the message contains.

		Note that Class.getMethod() is not used for method resolution.
		The getMethod() method will not match methods for which a
		parameter type is unknown. This algorithm will in the case that
		only one method name matches and has the same number of arguments.

		Instead the following algorithm is applied recursively to all
		this class and all superclasses until the Actor class is
		reached. Null arguments are given Object parameter types.

		1.) Check for an exact match between method name and parameter
		    type class names.

		2.) Check for an exact match between method name, parameter
		    number and for assignability between parameter types.

		3.) If above two checks fail, repeat for superclass if Actor
		    is the superclass then fail.
	*/
	public void process(Message message) {
		String methodName = message.getMethodName();
		Object[] messageArguments = message.getArguments();
		if (System.getProperty( "debug") != null) {
			String out="PROCESS: " + message.getMethodName() + "\t";
			for (int i=0; i < messageArguments.length; i++) {
				out += messageArguments[i]+" ";
			}
			System.err.println(out);
		}
		// Determine the parameter types for supplied arguments.
		//mpt stands for MessageParameterTypes,  i made this change for readability
		//and so the following code isn't 300 lines long
		//	--Travis
		Class[] mpt = new Class[ messageArguments.length ];

		for (int i = 0; i < mpt.length; ++i) {
			// Check to see if the argument is null. If so set its type
			// to be object.
			if (messageArguments[i] == null) {
				mpt[i] = null;
			} else {
				// Otherwise get the argument's type.
				mpt[i] = messageArguments[i].getClass();
			}
		}

		// Check the current Class and then all super classes for a matching method.
		Class state = getClass();
		Method match = null;

		while (!state.getName().equals( "java.lang.Object" )) {
			// Examine each method declared in this class.
			Method[] declared = state.getDeclaredMethods();

			for (int i = 0; i < declared.length; ++i) {
				Method candidate = declared[i];

				//cpt stands for candidateParameterTypes, for the same reason mpt stands for
				//messageParameterTypes
				Class[] cpt = candidate.getParameterTypes();

				// Check for an exact match between method name and parameter number.
				if (candidate.getName().equals( methodName ) && cpt.length == mpt.length) {
					// Check if all the parameter types match exactly.
					int j;
					for (j = 0; j < mpt.length; ++j) {
						if ( (mpt[j] == null) || !mpt[j].getName().equals( cpt[j].getName() )) {
							break;
                  				}
               				}

               				// Matched all successfully.
               				if (j == mpt.length) {
                  				match = candidate;
                 				break;
               				}
           			}
			}

         		// See if a match has been found at this class.
         		if (match != null) {
            			break;
         		}

         		for (int i = 0; i < declared.length; ++i) {
            			Method candidate = declared[i];

				//Again, cpt stands for CandidateParameterTypes, to save
				//valuable hard drive space
            			Class[] cpt = candidate.getParameterTypes();
            			// Check for an exact match between method name and parameter number.

            			if (candidate.getName().equals( methodName ) && cpt.length == mpt.length) {

               				// Check if the candidate parameter types are assignable from
               				// the message parameter types.
               				int j;
               				for (j = 0; j < mpt.length; ++j) {
                  				if ( !((mpt[j] == null) && !cpt[j].isPrimitive()) && !cpt[j].isAssignableFrom( mpt[j] ) ) {
							// If the candidate parameter type is a primitive type check to see if the
                     		// message parameter is any of the nine primitive wrapper classes
							// needs to be changed,  check and see if cpt is primitive and mpt is wrapper
							// If the candidate parameter type is a primitive type check to see if the
							// message parameter is any of the nine primitive wrapper classes

							if (!((cpt[j].isPrimitive() || cpt[j].isArray()) &&
							   (	cpt[j] == Integer.TYPE && mpt[j] == Integer.class ||
								cpt[j].getName().equals( "[I" ) && mpt[j].getName().equals( "[Ljava.lang.Integer;" ) ||

								cpt[j] == Boolean.TYPE && mpt[j] == Boolean.class ||
								cpt[j].getName().equals( "[Z" ) && mpt[j].getName().equals( "[Ljava.lang.Boolean;" ) ||

								cpt[j] == Character.TYPE && mpt[j] == Character.class ||
								cpt[j].getName().equals( "[C" ) && mpt[j].getName().equals( "[Ljava.lang.Character;" ) ||

								cpt[j] == Byte.TYPE && mpt[j] == Byte.class ||
								cpt[j].getName().equals( "[B" ) && mpt[j].getName().equals( "[Ljava.lang.Byte;" ) ||

								cpt[j] == Short.TYPE && mpt[j] == Short.class ||
								cpt[j].getName().equals( "[S" ) && mpt[j].getName().equals( "[Ljava.lang.Short;" ) ||

								cpt[j] == Long.TYPE && mpt[j] == Long.class ||
								cpt[j].getName().equals( "[J" ) && mpt[j].getName().equals( "[Ljava.lang.Long;" ) ||

								cpt[j] == Float.TYPE &&	mpt[j] == Float.class ||
								cpt[j].getName().equals( "[F" ) && mpt[j].getName().equals( "[Ljava.lang.Float;" ) ||

								cpt[j] == Double.TYPE && mpt[j] == Double.class ||
								cpt[j].getName().equals( "[D" ) && mpt[j].getName().equals( "[Ljava.lang.Double;" ) ||

								cpt[j] == Void.TYPE && mpt[j] == Void.class
							   ))) {
								break;
							}

							// If the matching candidate parameter is a primitive array type and
							// the message parameter is a wrapper type convert the message parameter
							// to a primitive array.
							if (cpt[j].isArray() && mpt[j].getName().equals( "[Ljava.lang.Integer;" )) {

								// Copy the values.
								int[] array = new int[ ((Integer[])messageArguments[j]).length ];
								for (int w=0; w < ((Integer[])messageArguments[j]).length; w++) {
									array[w] = ((Integer[])messageArguments[j])[w].intValue();
								}

								messageArguments[j] = array;
							}
							else if (cpt[j].isArray() && mpt[j].getName().equals( "[Ljava.lang.Boolean;" )) {

								// Copy the values.
								boolean[] array = new boolean[ ((Boolean[])messageArguments[j]).length ];
								for (int w=0; w < ((Boolean[])messageArguments[j]).length; ++w) {
									array[w] = ((Boolean[])messageArguments[j])[w].booleanValue();
								}

								messageArguments[j] = array;
							}
							else if (cpt[j].isArray() && mpt[j].getName().equals( "[Ljava.lang.Character;" )) {

								// Copy the values.
								char[] array = new char[ ((Character[])messageArguments[j]).length ];
								for (int w=0; w < ((Character[])messageArguments[j]).length; ++w) {
									array[w] = ((Character[])messageArguments[j])[w].charValue();
								}

								messageArguments[j] = array;
							}
							else if (cpt[j].isArray() && mpt[j].getName().equals( "[Ljava.lang.Byte;" )) {

								// Copy the values.
								byte[] array = new byte[ ((Byte[])messageArguments[j]).length ];
								for (int w=0; w < ((Byte[])messageArguments[j]).length; ++w) {
									array[w] = ((Byte[])messageArguments[j])[w].byteValue();
								}

								messageArguments[j] = array;
							}
							else if (cpt[j].isArray() && mpt[j].getName().equals( "[Ljava.lang.Short;" )) {

								// Copy the values.
								short[] array = new short[ ((Short[])messageArguments[j]).length ];
								for (int w=0; w < ((Short[])messageArguments[j]).length; ++w) {
									array[w] = ((Short[])messageArguments[j])[w].shortValue();
								}

								messageArguments[j] = array;
							}
							else if (cpt[j].isArray() && mpt[j].getName().equals( "[Ljava.lang.Long;" )) {

								// Copy the values.
								long[] array = new long[ ((Long[])messageArguments[j]).length ];
								for (int w=0; w < ((Long[])messageArguments[j]).length; ++w) {
									array[w] = ((Long[])messageArguments[j])[w].longValue();
								}

								messageArguments[j] = array;
							}
							else if (cpt[j].isArray() && mpt[j].getName().equals( "[Ljava.lang.Float;" )) {

								// Copy the values.
								float[] array = new float[ ((Float[])messageArguments[j]).length ];
								for (int w=0; w < ((Float[])messageArguments[j]).length; ++w) {
									array[w] = ((Float[])messageArguments[j])[w].floatValue();
								}

								messageArguments[j] = array;
							}
							else if (cpt[j].isArray() && mpt[j].getName().equals( "[Ljava.lang.Double;" )) {

								// Copy the values.
								double[] array = new double[ ((Double[])messageArguments[j]).length ];
								for (int w=0; w < ((Double[])messageArguments[j]).length; ++w) {
									array[w] = ((Double[])messageArguments[j])[w].doubleValue();
								}

								messageArguments[j] = array;
							}
						}
					}

					// Matched all successfully.
					if (j == mpt.length) {
						match = candidate;
						break;
					}
				}
			}

			// See if a match has been found at this class.
			if (match != null) {
				break;
			}

			// Examine the parent class next.
			state = state.getSuperclass();
		}

		// If a match was found attempt to invoke the method, otherwise generate an exception.
		Object token = null;

		if (match != null) {
			try {
				token = match.invoke( this, messageArguments );
			} catch (InvocationTargetException e) {
				Throwable wrapped = e.getTargetException();

				if ( !(wrapped instanceof CurrentContinuationException) ) {
					System.err.println( "Invocation of " + methodName + " generated a " +
							    wrapped.getClass().getName() + " exception" + wrapped.getMessage() );

					wrapped.printStackTrace();
				}
			} catch (IllegalAccessException e) {
				System.err.println( "Illegal access exception generated by the invocation of " + methodName + " " + e.getMessage() );

				e.printStackTrace();
			}
		} else {
			String detail = "Actor.process error:\n\t";

			detail += "Could not find a matching method in " + getClass().getName() +
				  " and its super classes for " + methodName + "(";

			for (int i=0;i<messageArguments.length;++i) {
				if (messageArguments[i] != null) detail += " " + messageArguments[i];
				else detail += " null";
			}
			System.err.println( detail + " )" );
			System.err.println("\tsent by: " + message.getSender().toString());
			System.err.println("\tsent to: " + message.getTarget().toString());
		}
		

		// Send the messages generated by the invoked message
		transportService.send( messages );
		messages = new Vector();

		// Send the continuation.
		if (currentContinuation != null) {
			//message.getTarget is a reference to this actor, and is needed
			//to send the return value as a "resolveToken" message to its
			//target actor(s).
			currentContinuation.resolve(message.getTarget(), token);
		}
		RunTime.finishedProcessingMessage();
	}
}
