/**
 * $Id: ActorReference.java,v 1.19 2003/09/09 22:25:13 deselt Exp $
 *
 * This is the main Actor class.
 */

package salsa.language;
import java.io.IOException;

import salsa.language.exceptions.SalsaException;
import salsa.language.exceptions.SalsaWrapperException;
import salsa.messaging.TheaterService;
import salsa.messaging.TransportService;
import salsa.naming.NamingService;
import salsa.naming.UAL;
import salsa.naming.UAN;
import salsa.naming.exceptions.MalformedUANException;
import salsa.naming.exceptions.MalformedUALException;

public class ActorReference implements java.io.Serializable {
	public UAN uan;
	public UAL ual;

	private transient TheaterService theaterService;
	private transient TransportService transportService;
	private transient NamingService namingService;

	/*
	 * These new constructors have replaced getReferenceByName and getReferenceByLocation
	 * They prevent messy actor creation and syntactically make more sense.
	 * By default, the compiler uses the ActorReference(UAL) constructor, as all actorreferneces 
	 * begin as anonymous actors, and become non-anonymous when they are bound to a UAN
	 */

	public ActorReference() {
		theaterService = ServiceFactory.getTheater();
		transportService = ServiceFactory.getTransport();
		namingService = ServiceFactory.getNaming();

		uan = null;
		ual = null;
	}

	public ActorReference(UAN _uan) {
		this();
		uan = _uan;
		ual = null;
	}

	public ActorReference(UAL _ual) {
		this();
		uan = null;
		ual = _ual;
	}


	/*
	 * These methods retrieve the UAN and UAL of the actor
	 */
	public UAN getUAN() {
		return uan;
	}

	public UAL getUAL() {
		if ( (uan != null) && (theaterService.getEntry(uan) instanceof UAL)) {
			return (UAL)theaterService.getEntry(uan);
		} else {
			return ual;
		}
	}


	/* 
	 * The following cover local and remote message sending for Actors
	 */
	public void send(Message message) {
		//If this actor has been sent a bind message, its UAN needs to be updated
		//so it can be serialized and access its state on the remote theater,  move the
		//states registry entry to the non-anonymous registry, and update the
		//anonymous registry with the uan

		//if the actor is local, get its state and put the message
		//in its mailbox.
		
		Actor ref = getActor();
		try {
			synchronized(ref) {
				ref.send(message);
			}
		} catch (NullPointerException npe) {
			//remote message send:
			//if the actor is non-anonymous(it has a UAN), get its UAL from the registry
			//if not, remote send the message to the specified UAL
			if (uan == null) {
				transportService.send(message, ual);
			} else {
				Object entry = theaterService.getEntry(uan);

				UAL remoteLocation;
				if (entry instanceof UAL) {
					remoteLocation = (UAL)entry;
				} else {
					remoteLocation = (UAL)ServiceFactory.getNaming().resolve(uan);
					theaterService.setEntry(uan, remoteLocation);
				}
				transportService.send(message, remoteLocation);
			}
		}
	}

	/*
	 * This retrives the actor from whichever registry it resides in
	 * if the actor is not anonymous (it has a UAN), it will get the
	 * state from the non-anonymous registry.  If the actor is foreign,
	 * this registry returns the apropriate ual
	 * If the actor is anonymous, it will get the state from the anonymous
	 * registry. If that actor has since become non-anonymous (through the bind method)
	 * that registry will return the UAN, and this reference will become non-anonymous
	 */
	public Actor getActor() {
		Object state = null;

		if (uan != null) {
			state = theaterService.getEntry(uan);
		} else if (ual != null) {
			state = theaterService.getEntry(ual);

			if (state instanceof UAL) {
				ual = (UAL) state;
				state = getActor();
			} else if (state instanceof UAN) {
				uan = (UAN) state;
				state = theaterService.getEntry(uan);
			}
		}

		if (state instanceof Actor) return (Actor)state;
		else return null;
	}



	/*
	 * Methods for printing out the identity of an actor
	 */
	public String toString() {
		String name = "";
		if (uan != null) name += uan.toString() + ", ";
		else name += "null, ";
		if (ual != null) name += ual.toString();
		else name += "null";
		return name;
	}

	public String getName() {
		return getClass().getName();
	}

	public boolean equals(Object o) {
		if (o instanceof ActorReference) {
			ActorReference ref = (ActorReference)o;
			//make sure all the references on these actors
			//are correct
			ref.getActor();
			getActor();

			if (uan != null && ref.uan != null) return uan.equals(ref.uan);
			else if (ual != null && ref.ual != null) return ual.equals(ref.ual);
		}
		return false;
	}

	/*
	 * Methods for Actor serializability
	 */
	private void writeObject(java.io.ObjectOutputStream out) throws SalsaException {
		//sending reference to local anonymous actor
		//need to keep the theater alive now since the reference has been
		//sent to the rest of the theater system.

		if (uan == null) {
			Object state = theaterService.getEntry(ual);

			if (state instanceof UAL) {
				ual = (UAL)state;
				state = theaterService.getEntry(ual);
			}
			if (state instanceof UAN) {
				uan = (UAN)state;

				state = theaterService.getEntry(uan);
				if (state instanceof UAL) ual = (UAL)state;
			}
		}

		if (uan == null) RunTime.receivedUniversalActor();

		try {
			out.writeObject(uan);
			out.writeObject(ual);
			out.flush();
		} catch (IOException e) {
			throw new SalsaWrapperException( "Exception during actor reference serialization.", e);
		}
	}

	private void readObject(java.io.ObjectInputStream in) throws SalsaException {
		try {
			uan = (UAN)in.readObject();
			ual = (UAL)in.readObject();
		} catch (Exception e) {
			throw new SalsaWrapperException( "Exception during actor reference deserialization.", e);
		}

		//get new instances of the services used by the Actor Reference
		theaterService = ServiceFactory.getTheater();
		transportService = ServiceFactory.getTransport();
		namingService = ServiceFactory.getNaming();

		if (uan == null) {
			Object entry = theaterService.getEntry(ual);
			if (entry instanceof UAN) uan = (UAN)entry;
		}
	}
}
