/**
 * This is the Message class.
 */

package salsa.language;

public class Message implements java.io.Serializable {
	/**
		Actor that the message is being sent from.
	*/
	private ActorReference sender;
	public	ActorReference getSender()	{ return sender; }

	/**
		Actor that the message is being sent to.
	*/
	private ActorReference target;
	public	ActorReference getTarget()	{ return target; }

	/**
		Name of the method that should be invoked.
	*/
	private String methodName;
	public	String getMethodName()		{ return methodName; }

	/**
		Array of values that are copied to be the arguments of the method. If one of
		the arguments is the value of a yet to be evaluated token an instance of the
		Token class should be subituted for it until it is evaluated.
	*/
	protected Object[] 	arguments;
	public	  Object[]	getArguments()	{ return arguments; }
	public 	  void 		setArguments(Object[] arguments) { this.arguments = arguments; }


	/**
		This stores where the return value from this message
		should be sent to (if anywhere).
		getTokenInput is used by the actor.process method to
		determine what the currentContinuation is.  
	*/
	private	TokenInput	output = null;
	public	TokenInput	getOutput()	{ return output; }

	/**
		This is the number of tokens which must be received by the message before
		it can be processed.
	*/
	private int		requiredTokens = 0;
	public	void		setRequiredTokens(int requiredTokens)	{ this.requiredTokens += requiredTokens; }
	public	int		getRequiredTokens()			{ return requiredTokens; }

	/**
		This is the synchronization token that the message waits for.
	*/
	private TokenOutput	input = null;
	public	TokenOutput	getInput()	{ return input; }
	public	void		setInput (Object o) {
		if (o == null) {
			this.input = null;
		} else if (o instanceof TokenOutput) {
			this.input = (TokenOutput)o;
			requiredTokens++;
		}
	}

	private String propertyName = null;
	private Object[] propertyParameters = null;
	public	void setProperty(String propertyName, Object[] propertyParameters) {
		this.propertyName = propertyName;
		this.propertyParameters = propertyParameters;
	}
	public	String getProperty() { return propertyName; }
	public	Object[] getPropertyParameters() { return propertyParameters; }


	/**
		This is the default constructor for messages.  TokenOutput is the synchronization token
		that the message waits for (there are also TokenOutput's within the arguments which
		the message will also wait for).  The TokenInput stores where the return value from this
		message will go to after it is finished being processed.
	*/
	public Message(ActorReference sender, ActorReference target, String methodName, TokenInput output) {
		this.sender = sender;
		this.target = target;
		this.methodName = methodName;
		this.output = output;
	}

	/**
		This indicates than an incoming token has been sent to this actor,
		and it has been resolved.
	 */
	public void receivedToken() {
		requiredTokens--;

		//requiredTokens equaling zero indicates that this message is
		//ready to be processed by its actor, so we send the message
		//to the actor.
		if (requiredTokens == 0) {
			if (System.getProperty( "debug") != null)
				System.err.println("SELFSENDING: " + getName());
			target.send(this);

			RunTime.finishedProcessingMessage();
		}
	}


	/**
		This method returns a human readable representation of
		what is in this message.  Used for debugging purposes.
	*/
	public String getName() {
		String text = "[Source: ";
		if (sender != null) text += sender.getUAL() + "(" + sender.getUAN() + ")";
		else text += "null";

		text += ", Target: ";
		if (target != null) text += target.getUAL() + "(" + target.getUAN() + ")";
		else text += "null";

		text += ", Method: " + methodName;

		if (arguments != null) {
			text += ", Arguments[" + arguments.length +"]: {";
			for (int i = 0; i < arguments.length; i++) {
/*				if (arguments[i] != null) text += arguments[i].getClass().getName();
				else text += "null";
*/
				text+=arguments[i];
				if (i != (arguments.length-1)) text += ", ";

			}
			text += "}";
		}
		else text += null;

		text += "]";
		return text;
	}
}
