package salsa.language;

import salsa.language.exceptions.*;
import salsa.naming.NamingService;
import salsa.messaging.*;

/**
 * The ServiceFactory manufactures references to the standard location dependent
 * services used by an actor. These include the TheaterService,
 * TransportService, NamingService, and ReceptionService.
 * 
 * In the case of a standalone actor, the ServiceFactory will create a theater
 * instance if one is needed. If a theater is already running a reference to
 * will be passed to the requesting actor.
 * 
 * The ServiceFactory is implemented as a synchronized, passive, singleton.
 * 
 * The default services are:<br>
 * <table> 
 * <tr><td>Default: </td><td>System Property: </td></tr>
 * <tr><td>wwc.messaging.Theater</td><td></td>-Dtheater=yourclass</tr>
 * <tr><td>wwc.messaging.Messenger</td><td>-Doutgoing=yourclass</td></tr>
 * <tr><td>wwc.messaging.Handler</td><td>-Dincoming=yourclass</td></tr>
 * <tr><td>wwc.naming.WwcNamingImpl</td><td>-Dnaming=yourclass</td></tr>
 * </table>
 * 
 * Classes defined in system properties must have no argument contructors and
 * should implement the their respective service interface.
 * 
 * @author stepha
 */

public class ServiceFactory {
	// Service implementations.
	private static TheaterService theater = null;
	private static TransportService outgoing = null;
	private static ReceptionService incoming = null;
	private static NamingService naming = null;

	/**
	 * Constructor for Service Factory
	 * This allows for different service implementations
	 * to be specified.
	 */
	public synchronized static void createFactory(TheaterService _theater, TransportService _outgoing, 
						      ReceptionService _incoming, NamingService _naming) {
		theater = _theater;
		incoming = _incoming;
		outgoing = _outgoing;
		naming = _naming;
	}


	/**
	 * Returns an implementation of the transport service. If no service has been specified,
	 * it checks the system property "-Doutgoing=<transportservice>" and instantiates that
	 * implementation
	 * @return TransportService
	 */
	public synchronized static TransportService getTransport() {
		if (outgoing == null) {
			// Check for the messagingService.
			String className = System.getProperty( "outgoing" );
			if (className == null) {
				className = "wwc.messaging.Messenger";
			}

			try {
				outgoing = (TransportService)Class.forName( className ).newInstance();

			} catch (Exception e) {
				throw new SalsaWrapperException( "Exception creating default transport service.", e );
			}
		}
		
		return outgoing;
	}

	/**
	 * Returns the ReceptionService. If no reception service is specified,
	 * it uses the system property "-Dincoming=<receptionservice>" to
	 * instantiate the given service.
	 * @return ReceptionService
	 */
	public synchronized static ReceptionService getReception() {
		if (incoming == null) {
			// Check for the migration service.
			String className = System.getProperty( "incoming" );
			if (className == null) {
				className = "wwc.messaging.Handler";
			}

			try {
				incoming = (ReceptionService)Class.forName( className ).newInstance();
			} catch (Exception e) {
				throw new SalsaWrapperException( "Exception creating default reception service.", e );
			}
		}
		
		return incoming;
	}

	/**
	 * Returns the namingImpl.
	 * @return NamingService
	 */
	public synchronized static NamingService getNaming() {
		if (naming == null) {
			// Check for the naming service.
			String className = System.getProperty( "naming" );
			NamingService MessagingServiceImpl;
			if (className == null) {
				className = "wwc.naming.WwcNamingImpl";
			}

			try {
				naming = (NamingService)Class.forName( className ).newInstance();

			} catch (Exception e) {
				throw new SalsaWrapperException( "Exception creating default naming service.", e );
			}
		}
			
		return naming;
	}

	/**
	 * Returns the theaterImpl.
	 * @return TheaterService
	 */
	public synchronized static TheaterService getTheater() {
		if (theater == null) {
			String className = System.getProperty( "theater" );
			if (className == null) {
				className = "wwc.messaging.Theater";
			}

			try {
				theater = (TheaterService)Class.forName( className ).newInstance();
			} catch (Exception e) {
				throw new SalsaWrapperException( "Exception creating default theater service.", e );
			}
		}

		return theater;
	}
}
